from fastapi import WebSocketDisconnect
import os
import base64
import json
import asyncio
import websockets
import pyaudio
import time
import numpy as np
import wave
import uuid
import myopenai
import queue
from io import BytesIO


class myopenai_realtime:

    def __init__(self, 
            api_key, 
            client_ws, 
            f_chatmode:bool             = False, 
            f_whispermode:bool          = False,
            model:str                   = 'gpt-4o-realtime-preview-2024-12-17',
            AUDIO_SEND_SIGNAL:str       = 'input_audio_buffer.append', #クライアントからの音声を受信するシグナル
            SILENCE_THRESHOLD:int       = 500, # 無音と判断する閾値
            SILENCE_DURATION_LIMIT:float= 1.0 # 無音と判断する時間(s)
    ):
        
        self.f_client_socket_connected = False #クライアントとの接続があるかどうか
        self.client_id = str(uuid.uuid4()) #ランダム生成
        self.client_ws = client_ws
        self.api_key   = api_key
        self.model     = model
        self.AUDIO_SEND_SIGNAL = AUDIO_SEND_SIGNAL
        self.f_responding = False #OpenAIがレスポンスを返しているときにOnになる
        self.f_playing = False #クライアントがレスポンスを再生してたらTrue
        self.audio_chunks = [] #ユーザーのマイク音が入る（OpenAIに送られた分だけ）
        self.mo = myopenai.myopenai()

        # 無音と判断する閾値（PCM16の絶対値平均などで判断）
        self.SILENCE_THRESHOLD = SILENCE_THRESHOLD # 雑音量。調整が必要
        if f_chatmode :
            self.SILENCE_DURATION_LIMIT = max(SILENCE_DURATION_LIMIT, 1.0) # 1秒間無音が続いた場合
        else :
            self.SILENCE_DURATION_LIMIT = SILENCE_DURATION_LIMIT # 1秒間無音が続いた場合

        

        self.last_sound_time = time.time()
        self.audiodata_comming = True
        self.f_check_by_gpt = False
        self.dic_temptalk = {}
        self.l_messages_whisper = [] #[{id:"1", text:"こんにちは"},{id:2, text:"ありがとうございます"}]
        self.f_whispermode = f_whispermode #ユーザーからのインプットを文字起こしする

        self.audio_queue = queue.Queue()  # 無音で追加される。追加されたらWhisperが走る
        self.whisper_result_queue = queue.Queue() #Whisperの結果を入れるキュー。外部で監視
        self.f_whisper_running = False #Whisperが実行中かどうか

        self.l_messages = []
        self.audio_chunks_openai = []
        self.openai_ws = None
        self.f_chatmode = f_chatmode #ユーザーからのインプットをrealtime apiに流す

    #--------------------------------------------------------------
    # クライアントにIDを送信
    #--------------------------------------------------------------
    async def send_client_id(self):
        await self.send_to_client("send_client_id", self.client_id, "client_id")

    def get_client_id(self):
        return self.client_id

    #--------------------------------------------------------------
    # OpenAIインスタンスを取得（外で使いまわす）
    #--------------------------------------------------------------
    def get_mo_instance(self):
        return self.mo
    
    #--------------------------------------------------------------
    # Whisperが実行中かどうか
    #-------------------------------------------------------------- 
    def is_whisper_running(self):
        return self.f_whisper_running
    
    #--------------------------------------------------------------
    # クライアントにメッセージを送信
    #--------------------------------------------------------------
    async def send_to_client(self, type, text, keyname:str="text"):
        if self.client_ws is not None:
            await self.client_ws.send_text(json.dumps({
                "type" : type,
                keyname: text
            }))
            print("クライアントWebSocketでメッセージを送信しました。")
        else:
            print("クライアントWebSocket接続がありません")

    #--------------------------------------------------------------
    # クライアントとの接続があるかどうか
    #--------------------------------------------------------------
    def is_client_socket_connected(self):
        return self.f_client_socket_connected

    #--------------------------------------------------------------
    # クライアントからのメッセージ（音声）を受信
    #--------------------------------------------------------------
    async def handle_client_websocket(self):
        self.f_client_socket_connected = True

        # WebSocketの処理と並行してキューを監視
        asyncio.create_task(self.monitor_audio_queue())

        try:
            self.last_sound_time = time.time()
            while self.is_client_socket_connected():
                message = await self.client_ws.receive_text()
                data = json.loads(message)

                #マイクからの音声を受信
                if data.get("type") == self.AUDIO_SEND_SIGNAL and "audio" in data:
                    base64_audio = data["audio"]
                    audio_data = base64.b64decode(base64_audio)
                    audio_array = np.frombuffer(audio_data, dtype=np.int16)
                    amplitude = np.abs(audio_array).mean()
                    current_time = time.time()

                    # 無音検出(大きな音量がある場合、受信時間を更新)
                    if amplitude > self.SILENCE_THRESHOLD:
                        self.last_sound_time = current_time
                        self.audiodata_comming = True
                    else:
                        # 無音検出→Whisper実行
                        if current_time - self.last_sound_time >= self.SILENCE_DURATION_LIMIT:
                            self.audiodata_comming = False

                    # 有音検出→ため込み & OpenAIへの音声送信
                    if self.audiodata_comming:
                        base64_audio_utf8 = base64.b64encode(audio_data).decode("utf-8")
                        audio_event = {
                            "type": "input_audio_buffer.append",
                            "audio": base64_audio_utf8
                        }
                        # print("音声受信")
                        self.audio_chunks.append(audio_data)  # 音声データを蓄積

                        # OpenAIへの音声送信
                        if self.openai_ws is not None:
                            # 回答生成中や再生中は、音を送らない
                            if self.f_responding == False and self.f_playing == False :
                                if self.f_chatmode :
                                    await self.openai_ws.send(json.dumps(audio_event))
                                # print("sent audio data to openai")
                            else :
                                print("回答生成中もしくは再生中のため、OpenAIへの音声送信中止")
                        else:
                            if self.f_chatmode :
                                print("OpenAI接続がまだ確立されていません。")

                    await asyncio.sleep(0)




                elif data.get("type") == "status" and data["message"] == "done playing" :
                    # 終了処理 
                    self.f_playing = False
                    print("レスポンス再生終了")

                else:
                    # エラー処理
                    print("未知のメッセージタイプを受信:", data.get("type"))

                # 無音検出→Whisperキューにためる（実行は別の監視プログラム）
                if not self.audiodata_comming :
                    if self.f_whispermode and len(self.audio_chunks) > 10 : #「こんにちわ」で200くらいいく
                        print(f"無音検出→Whisper実行: {len(self.audio_chunks)}")
                        combined_audio = b''.join(self.audio_chunks)
                        self.save_wav_file(combined_audio, "user_talk.wav")
                        self.audio_queue.put(combined_audio) # キューにcombined_audioを追加(→Whisper実行)
                        self.audio_chunks = []

        except WebSocketDisconnect:
            self.f_client_socket_connected = False

        print("クライアントとの接続が切断されました。")
        await self.close_openai_ws()

    #--------------------------------------------------------------
    # キューを監視し、オーディオデータが来たらWhisperを実行
    #--------------------------------------------------------------
    async def monitor_audio_queue(self):
        while self.f_client_socket_connected:
            if not self.audio_queue.empty():
                audio_data = self.audio_queue.get()
                await self.mywhisper(audio_data)
            await asyncio.sleep(0.1)  # 適切な間隔で監視

        print("クライアントとの接続が切断されたので、オーディオキュー監視を終了します。")

    #--------------------------------------------------------------
    # Whisperで音声をテキストに変換
    #--------------------------------------------------------------
    async def mywhisper(self, all_audio_data):
        print("whisper start")
        self.f_whisper_running = True
        d_whisper_result = self.mo.speech_to_text_pcm(all_audio_data)
        self.whisper_result_queue.put(d_whisper_result)
        print(d_whisper_result)

        self.l_messages_whisper.append({"id": str(len(self.l_messages_whisper)+1), "text": d_whisper_result["text"]})
        await self.send_to_client("talk.add", self.l_messages_whisper[-1])

        d_cost = self.mo.get_cost_all()
        # with open("data/cost.json", "w", encoding="utf-8") as f:
        #     json.dump(d_cost, f, ensure_ascii=False, indent=4)
        
        self.f_whisper_running = False

    #--------------------------------------------------------------
    # Whisperの結果をキューから取得
    #--------------------------------------------------------------
    def get_whisper_queue(self) -> str :
        # {'text': 'こんにちは', 'duration': 0.949999988079071, 'segments': [{'text': 'こんにちは', 'start': 0.0, 'end': 2.559999942779541}]}
        txt = ""
        while not self.whisper_result_queue.empty() :
            d = self.whisper_result_queue.get(timeout=0.1)
            txt += f'{d["text"]}\n'
        return txt

    #--------------------------------------------------------------
    # WAVファイル保存
    #--------------------------------------------------------------
    def save_wav_file(self, all_audio_data, filename="output.wav"):
        p = pyaudio.PyAudio()
        try:
            with wave.open(filename, "wb") as wf:
                wf.setnchannels(1)  
                wf.setsampwidth(p.get_sample_size(pyaudio.paInt16))
                wf.setframerate(24000)
                wf.writeframes(all_audio_data)
            print(f"WAVファイルとして保存しました: {filename}")
        finally:
            p.terminate()


    # def create_wav_buffer(self, audio_data, f_save:bool=False):
    #     audio_buffer = BytesIO()
    #     p = pyaudio.PyAudio()
    #     try:
    #         with wave.open(audio_buffer, "wb") as wf:
    #             wf.setnchannels(1)
    #             wf.setsampwidth(p.get_sample_size(pyaudio.paInt16))
    #             wf.setframerate(24000)
    #             wf.writeframes(audio_data)
    #     finally:
    #         p.terminate()
        
    #     if f_save:
    #         filename = f"temp_{uuid.uuid4().hex[:8]}.wav"
    #         self.save_wav_file(audio_data, filename)

    #     audio_buffer.seek(0)
    #     return audio_buffer























    #-------------------------------------------------------------- 
    # OpenAIとのソケットを開く
    #--------------------------------------------------------------
    async def connect_and_session_update(self, instructions:str="", voice:str="echo"):
        WS_URL = f"wss://api.openai.com/v1/realtime?model={self.model}"
        HEADERS = {
            "Authorization": "Bearer "+ self.api_key, 
            "OpenAI-Beta": "realtime=v1"
        }
        self.openai_ws = await websockets.connect(WS_URL, extra_headers=HEADERS)
        print("OpenAI WebSocketに接続しました。")

        session_update = {
            "type": "session.update",
            "session": {
                "modalities": ["audio", "text"],
                "instructions": instructions,
                "voice": voice,
                "input_audio_transcription": {
                    "model": "whisper-1"
                },
                "turn_detection": {
                    "type": "server_vad"
                }
            }
        }
        await self.openai_ws.send(json.dumps(session_update))
        print("初期設定を送信しました。")
        asyncio.create_task(self.receive_openai_response())

        return {"status": "openai stream started"}

    #--------------------------------------------------------------
    # OpenAIとのソケットを閉じる
    #-------------------------------------------------------------- 
    async def close_openai_ws(self):
        if self.openai_ws is not None:
            await self.openai_ws.close()
            self.openai_ws = None
            print("OpenAIとの接続を切断")
            response_message = "OpenAIとのWebSocket接続を閉じました"
        else:
            response_message = "OpenAIとのWebSocket接続は存在しません"

        return {"message": response_message}

    #--------------------------------------------------------------
    # OpenAIからのレスポンスを受信
    #--------------------------------------------------------------
    async def receive_openai_response(self):
        txt_response = ""
        user_transcript = ""
        
        print("OpenAIからのレスポンスリスニング開始")
        l_temp_response = []
        while self.openai_ws is not None:
            try:
                response = await self.openai_ws.recv()

                # print(f"OpenAIからのレスポンスを受信")
            except websockets.exceptions.ConnectionClosed:
                print("OpenAIとのWebSocket接続が閉じられました。受信ループを終了します。")
                break
            except Exception as e:
                print(f"OpenAIとのWebSocket受信中にエラーが発生しました: {e}")
                break

            response_data = json.loads(response)
            l_temp_response.append(response_data)
            with open("data/temp_response.json", "w", encoding="utf-8") as f:
                json.dump(l_temp_response, f, ensure_ascii=False, indent=4)


            if "type" in response_data and response_data["type"] not in ['response.audio_transcript.delta', 'response.audio.delta']:
                print(f'msg from openai: {response_data["type"]}')

            if "type" in response_data:
                type = response_data["type"]
                if type == "session.created":
                    # print(response_data)
                    await self.client_ws.send_text(json.dumps({
                        "type": "session.created",
                        "session_id": response_data["session"]["id"]
                    }))
                elif type == "response.audio_transcript.delta":
                    self.f_playing = True
                    txt_response += response_data["delta"]
                    await self.client_ws.send_text(json.dumps({
                        "type": "response.assistant.delta",
                        "text": response_data["delta"]
                    }))
                elif type == "response.audio.delta":
                    base64_audio_response = response_data["delta"]
                    if base64_audio_response:
                        try:
                            self.audio_chunks_openai.append(base64.b64decode(base64_audio_response))
                            await self.client_ws.send_text(json.dumps({
                                "type": "response.audio",
                                "audio": base64_audio_response
                            }))
                        except Exception as e:
                            print(f"クライアント {self.client_id} への音声送信エラー: {e}")
                elif type == "response.audio_transcript.done":
                    print(f"---GPT said: {response_data['transcript']}")
                    print(f"---diff    : {txt_response} ")
                    self.l_messages.append({"type":"assistant", "text":response_data['transcript']})
                    await self.client_ws.send_text(json.dumps({
                        "type": "response.assistant",
                        "text": response_data['transcript']
                    }))
                elif type == "response.audio.done":
                    # print(f"whiper text = {self.whisper_text}")
                    # WAVファイルとして保存
                    all_audio_data = b"".join(self.audio_chunks_openai)
                    await asyncio.to_thread(self.save_wav_file, all_audio_data, "response_from_openai.wav")
                    self.audio_chunks_openai = []
                elif type == "conversation.item.input_audio_transcription.completed":
                    user_transcript = response_data["transcript"]
                    self.l_messages.append({"type":"user", "text":user_transcript})
                    print(f"---You said: {user_transcript}")

                    await self.client_ws.send_text(json.dumps({
                        "type": "response.user",
                        "text": user_transcript
                    }))
                elif type == 'input_audio_buffer.committed' :
                    print("インプット音声完成（送った音声をファイル保存）")
                    all_audio_data = b"".join(self.audio_chunks)
                    # WAVファイルとして保存
                    await asyncio.to_thread(self.save_wav_file, all_audio_data, "output.wav")      
                    # print("testtesttesttesttesttesttesttesttest")              
                    self.audio_chunks = []

                elif type == 'conversation.item.created' :
                    print("レスポンス生成開始！！")
                    self.f_responding = True
                    await self.client_ws.send_text(json.dumps({
                        "type": "isResponding",
                        "flag": True
                    }))
                elif type == "response.done":
                    usage = response_data["response"]["usage"]
                    d_cost = {
                        "model"                     : self.model,
                        "tokens_input"              : usage["input_token_details"]["text_tokens"],
                        "tokens_input_cached"       : usage["input_token_details"]["cached_tokens_details"]["text_tokens"],
                        "tokens_input_audio"        : usage["input_token_details"]["audio_tokens"],
                        "tokens_input_audio_cached" : usage["input_token_details"]["cached_tokens_details"]["audio_tokens"],
                        "tokens_output"             : usage["output_token_details"]["text_tokens"],
                        "tokens_output_audio"       : usage["output_token_details"]["audio_tokens"]
                    }
                    self.mo.l_cost.append(d_cost)
                    cost = self.mo.get_cost(d_cost)
                    print(f"----- cost: {cost:.2f} ドル")

                    txt_response = ""
                    user_transcript = ""
                    self.f_responding = False
                    print("レスポンス生成終了！！")
                    await self.client_ws.send_text(json.dumps({
                        "type": "response.done",
                        "text": "done"
                    }))
                elif type in ['response.content_part.done', 'response.output_item.done']:
                    pass

        print("OpenAIからのレスポンスリスニング終了")

    #--------------------------------------------------------------
    # OpenAIにメッセージを送信
    #-------------------------------------------------------------- 
    async def send_message(self, msg:str, role:str="user", voice:str="echo"):
        event = {
            "type": 'conversation.item.create',
            "item": {
                "type": 'message',
                "role": role,
                "content": [
                    {
                        "type": 'input_text',
                        "text": msg
                    }
                ]
            }
        }
        await self.openai_ws.send(json.dumps(event))

        init_request = {
            "type": "response.create",
            "response": {
                "modalities": ["audio", "text"],
                "voice": voice
            }
        }
        await self.openai_ws.send(json.dumps(init_request))
        print("初期リクエストを送信")
        return {"status": "xxx_success"}

